import datetime

from django.db import models
from django.db.models.query import QuerySet
from django.test import TestCase

from chainablemanager import ChainableManager


class BookManager(ChainableManager):
    class QuerySetMixin:

        def published_in(self, year):
            return self.filter(year=year)

        def in_genre(self, genre):
            return self.filter(genre=genre)


class Author(models.Model):
    name = models.CharField(max_length=255)


class Book(models.Model):
    name = models.CharField(max_length=255)
    author = models.ForeignKey(Author, related_name='books')
    genre = models.CharField(max_length=255)
    year = models.IntegerField()

    objects = BookManager()


def populate_books():
    mieville = Author.objects.create(name='China Mieville')
    banks = Author.objects.create(name='Iain Banks')

    mieville.books.create(name="The City & the City", genre="Crime", year=2009)
    mieville.books.create(name="Kraken", genre="Fantasy", year=2010)
    mieville.books.create(name="Embassytown", genre="Sci Fi", year=2011)
    mieville.books.create(name="Railsea", genre="Fantasy", year=2012)

    # To be pedantic, his Sci Fi works are published under  Iain M. Banks
    banks.books.create(name="The Business", genre="Fantasy", year=1999)
    banks.books.create(name="Dead Air", genre="Fantasy", year=2002)
    banks.books.create(name="The Algebraist", genre="Sci Fi", year=2004)
    banks.books.create(name="Matter", genre="Sci Fi", year=2008)
    banks.books.create(name="The Hydrogen Sonata", genre="Sci Fi", year=2012)
    banks.books.create(name="Stonemouth", genre="Fantasy", year=2012)


class ChainableManagerTests(TestCase):

    def setUp(self):
        populate_books()

    def test_simple_chain(self):

        railsea = Book.objects.get(name="Railsea")
        hydrogen_sonata = Book.objects.get(name="The Hydrogen Sonata")
        stonemouth = Book.objects.get(name="Stonemouth")

        self.assertTrue(hasattr(Book.objects, 'published_in'),
            "Model Manager has chainable method")
        self.assertTrue(callable(Book.objects.published_in),
            "Model Manager has callable chained method")

        books_2012 = Book.objects.published_in(2012)

        self.assertTrue(hasattr(books_2012, 'in_genre'),
            "QuerySet has chainable method")
        self.assertTrue(callable(books_2012.in_genre),
            "QuerySet has callable chained method")

        self.assertEqual(books_2012.count(), 3,
            "Three books published in 2012")
        self.assertEqual(
            set(books_2012), set([railsea, hydrogen_sonata, stonemouth]),
            "Got the correct three books")

        scifi_books_2012 = books_2012.in_genre('Sci Fi')

        self.assertEqual(scifi_books_2012.count(), 1,
            "One sci fi book published in 2012")
        self.assertEqual(
            set(scifi_books_2012), set([hydrogen_sonata]),
            "Got the correct book")

    def test_related_queries(self):
        mieville = Author.objects.get(name='China Mieville')
        banks = Author.objects.get(name='Iain Banks')

        hydrogen_sonata = Book.objects.get(name="The Hydrogen Sonata")
        stonemouth = Book.objects.get(name="Stonemouth")

        mievilles_books = mieville.books
        banks_books = banks.books

        self.assertTrue(hasattr(mievilles_books, 'published_in'),
            "Related Model Manager has chainable method")
        self.assertTrue(callable(mievilles_books.published_in),
            "Related Model Manager has callable chained method")

        self.assertEqual(mievilles_books.count(), 4,
            "We have the correct number of books")
        self.assertEqual(
            set(banks_books.published_in(2012)),
            set([hydrogen_sonata, stonemouth]),
            "Can access first chained method on related manager")
        self.assertEqual(
            set(banks_books.published_in(2012).in_genre("Fantasy")),
            set([stonemouth]),
            "Can access second chained method on related manager")

    def test_correct_classes(self):

        manager = Book.objects

        self.assertIsInstance(manager, models.Manager,
            "Manager should be an instance of models.Manager")
        self.assertIsInstance(manager, ChainableManager,
            "Manager should be an instance of ChainableManager")
        self.assertIsInstance(manager, manager.QuerySetMixin,
            "Manager should be an instance of its QuerySetMixin")

        queryset = manager.all()

        self.assertIsInstance(queryset, QuerySet,
            "QuerySet should be an instance of QuerySet")
        self.assertIsInstance(queryset, manager.ChainableQuerySet,
            "QuerySet should be an instance of manager.ChainableQuerySet")
        self.assertIsInstance(queryset, manager.QuerySetMixin,
            "QuerySet should be an instance of its QuerySetMixin")

    def test_correct_related_classes(self):

        manager = Book.objects

        mieville = Author.objects.get(name='China Mieville')
        mievilles_books = mieville.books

        self.assertIsInstance(mievilles_books, models.Manager,
            "Related manager should be an isntance of models.Manager")
        self.assertIsInstance(mievilles_books, ChainableManager,
            "Related manager should be an instance of manager.ChainableQuerySet")
        self.assertIsInstance(mievilles_books, manager.QuerySetMixin,
            "Related manager should be an instance of its QuerySetMixin")

        queryset = mievilles_books.all()

        self.assertIsInstance(queryset, QuerySet,
            "Related QuerySet should be an instance of QuerySet")
        self.assertIsInstance(queryset, manager.QuerySetMixin,
            "Related QuerySet should be an instance of its QuerySetMixin")



from functools import wraps

from django.db.models import Manager
from django.db.models.query import QuerySet
from django.utils.six import with_metaclass


def _make_proxy(name, fn):
    @wraps(fn)
    def _proxy(self, *args, **kwargs):
        qs = self.get_queryset()
        return getattr(qs, name)(*args, **kwargs)

    return _proxy


class ChainableManagerMetaclass(type(Manager)):
    """
    Metaclass for ChainableManager.
    """
    def __new__(cls, name, bases, attrs):
        # Construct the class as normal so we can examine it. We will not
        # actually use this class, unless there is no queryset mixin
        temp_cls = super(ChainableManagerMetaclass, cls).__new__(
            cls, name, bases, attrs)

        # Get the custom QuerySet mixin defined on the class
        QuerySetMixin = getattr(temp_cls, 'QuerySetMixin', None)

        # Bail here if there is no QuerySetMixin
        if QuerySetMixin is None:
            return temp_cls

        # Make a custom QuerySet from the mixin
        ChainableQuerySet = type(
            'ChainableQuerySet', (QuerySetMixin, QuerySet), {})

        # Make a new class with the mixin in place
        attrs['ChainableQuerySet'] = ChainableQuerySet
        bases = bases + (QuerySetMixin, )

        cls = super(ChainableManagerMetaclass, cls).__new__(
            cls, name, bases, attrs)

        return cls


class ChainableManager(with_metaclass(ChainableManagerMetaclass, Manager)):
    """
    A Model Manager that allows chaining custom filters and other methods on
    both the manager and any querysets produced by it.

    Add a class named `QuerySetMixin` to the Manager, and define all your
    custom, chainable methods on this class instead.
    """

    use_for_related_fields = True

    def get_queryset(self):
        """
        Create a QuerySet for querying this model. Will also have all the
        chainable methods defined on `QuerySetMixin`.
        """
        return self.ChainableQuerySet(self.model, using=self._db)



#!/usr/bin/env python
"""
Install django-chainable-manager using setuptools
"""

from setuptools import setup, find_packages

from chainablemanager import __version__

with open('README.rst', 'r') as f:
    readme = f.read()

setup(
    name='django-chainable-manager',
    version=__version__,
    description='Chainable methods on Model managers',
    long_description=readme,
    author='Tim Heap',
    author_email='tim@timheap.me',
    url='https://bitbucket.org/tim_heap/django-chainable-manager',

    install_requires=['Django>=1.4'],
    zip_safe=False,

    packages=find_packages(),

    include_package_data=True,
    package_data={},

    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Framework :: Django',
    ],
)
